/* $Id: hipe_x86_glue.h,v 1.4 2001/06/01 11:42:50 mikpe Exp $
 * hipe_x86_glue.h
 */

extern unsigned x86_call_to_native(Process*);
extern unsigned x86_return_to_native(Process*);
extern unsigned x86_tailcall_to_native(Process*);
extern unsigned x86_throw_to_native(Process*);
extern void nbif_fail(void);
extern void nbif_return(void);
extern void nbif_closure_call_emu(void);

static Eterm x86_catch_fail;

static __inline__ void hipe_arch_glue_init(void)
{
    unsigned i;

    if( (unsigned long)nbif_return & 0x3 )	/* sanity check */
	abort();
    i = beam_catches_cons((uint32*)nbif_fail, BEAM_CATCHES_NIL);
    x86_catch_fail = make_catch(i);
}

static __inline__ void hipe_push_x86_trap_frame(Process *p)
{
    /* ra is already saved in p->nsp[0]; just push catch frame */
    *--(p->nsp) = x86_catch_fail;
}

static __inline__ void hipe_pop_x86_trap_frame(Process *p)
{
    ASSERT(p->nsp[0] == x86_catch_fail);
    ++(p->nsp);
    /* ra is now in p->nsp[0] */
}

static __inline__ void
hipe_push_x86_params(Process *p, unsigned arity, Eterm reg[])
{
    Eterm *nsp;
    unsigned i;

    nsp = p->nsp;
    for(i = 0; i < arity; ++i)
	*--nsp = reg[i];
    p->nsp = nsp;
}

static __inline__ void
hipe_pop_x86_params(Process *p, unsigned arity, Eterm reg[])
{
    Eterm *nsp;
    int i;

    nsp = p->nsp;
    for(i = arity; --i >= 0;)
	reg[i] = *nsp++;
    p->nsp = nsp;
}

static __inline__ unsigned
hipe_call_to_native(Process *p, unsigned arity, Eterm reg[])
{
    /* Note that call_to_native() needs two words on the stack:
       one for the nbif_return return address, and one for the
       callee's return address should it need to call inc_stack_0. */
    hipe_check_nstack(p, 1+arity+1+1);
    hipe_push_x86_trap_frame(p);		/* needs 1 word */
    hipe_push_x86_params(p, arity, reg);	/* needs arity words */
    return x86_call_to_native(p);		/* needs 1+1 words */
}

static __inline__ unsigned
hipe_tailcall_to_native(Process *p, unsigned arity, Eterm reg[])
{
    hipe_check_nstack(p, arity+1);	/* +1 so callee can call inc_stack_0 */
    if( arity ) {
	Eterm nra;
	nra = *(p->nsp++);
	hipe_push_x86_params(p, arity, reg);
	*--(p->nsp) = nra;
    }
    return x86_tailcall_to_native(p);
}

static __inline__ void hipe_return_from_native(Process *p)
{
    hipe_pop_x86_trap_frame(p);
}

static __inline__ void hipe_throw_from_native(Process *p) { }

static __inline__ int
hipe_call_from_native_is_recursive(Process *p, Eterm reg[])
{
    Eterm nra;

    nra = *(p->nsp++);
    hipe_pop_x86_params(p, p->arity, reg);
    if( nra != (Eterm)nbif_return ) {
	*--(p->nsp) = nra;
	return 1;
    }
    hipe_pop_x86_trap_frame(p);
    return 0;
}

static __inline__ unsigned hipe_return_to_native(Process *p)
{
    return x86_return_to_native(p);
}

static __inline__ unsigned hipe_throw_to_native(Process *p)
{
    return x86_throw_to_native(p);
}

static __inline__ void hipe_reschedule_from_native(Process *p)
{
    ASSERT(p->arity == 0);
}

static __inline__ unsigned
hipe_reschedule_to_native(Process *p, unsigned arity, Eterm reg[])
{
    ASSERT(arity == 0);
    return x86_tailcall_to_native(p);
}

/* Return the address of a stub switching a native closure call to BEAM. */
static __inline__ void *hipe_closure_stub_address(unsigned arity)
{
    return nbif_closure_call_emu;
}
