%%%----------------------------------------------------------------------
%%% File    : hipe_alias.erl
%%% Author  : Ingemar berg <d95ina@zeppo.it.uu.se>
%%% Purpose : Keep track of the aliases created by the coalescing.
%%% Created : 18 Mar 2000 by Ingemar berg <d95ina@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_alias).
-export([initAlias/1, getAlias/3, setAlias/3]).

%%%----------------------------------------------------------------------
% Function:     initAlias
%
% Description:  Initialize an alias mapping with NrOfNodes slots.
% Parameters:
%   NrOfNodes      -- The number of registers that might need an alias.
%
% Returns:
%  Alias           -- The initial alias mapping.
%
%%%----------------------------------------------------------------------

initAlias(NrOfNodes) ->
    {alias, hipe_vectors:init(NrOfNodes)}.

%%%----------------------------------------------------------------------
% Function getAlias
%
% Description: Get the alias of a node.
% Parameters:
%   Node           -- The node, whose alias shall be found 
%   NodeSets       -- A structure containing the different NodeSets
%   Alias          -- The alias mapping
%
% Returns:
%   Name           -- The requested alias
%
%%%----------------------------------------------------------------------

getAlias(Node,NodeSets,{alias, Alias}) when integer(Node) ->
    case hipe_node_sets:member(coalesced,Node,NodeSets) of 
	true ->
	    getAlias(hipe_vectors:get(Alias,Node),NodeSets,{alias, Alias});
	false ->
	    Node
    end;
getAlias(Node,NodeSets,{alias, Alias}) ->
    hipe_message:error(getAlias, "Node not integer: ~p", [Node]).

%%%----------------------------------------------------------------------
% Function: setAlias
%
% Description: Set the alias of a node.
% 
% Parameters:
%   Node           -- The node that shall get an alias
%   ToNode         -- The new name for the node
%   Alias          -- The alias mapping the info shall be stored in
%
% Returns:
%   Alias          -- The new alias mapping
%
%%%----------------------------------------------------------------------

setAlias(Node,ToNode,{alias, Alias}) when integer(Node), integer(ToNode) ->
    {alias, hipe_vectors:set(Alias,Node,ToNode)};
setAlias(Node,ToNode,{alias,Alias}) ->
    hipe_message:error(setAlias, "Node not integer: ~p or ~p", [Node,ToNode]).
