%% Copyright (c) 1998 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <01/02/12 13:45:27 happi>
%% ====================================================================
%%  Filename : 	patch.erl
%%  Module   :	patch
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 1998-01-15 Erik Johansson (happi@csd.uu.se): Created.
%% CVS:
%%    $Author: richardc $
%%    $Date: 2001/03/26 18:37:08 $
%%    $Revision: 1.1.1.1 $
%% ====================================================================
%% Exported functions (short description):
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_patch).
-export([patch/2]).


patch([{Address,Refs}|Patchlist],Dest) ->
  patch_refs(callees_adt:to_list(Refs),Address,Dest),
  patch(Patchlist,Dest);
patch([],_) -> true.


patch_refs([Ref|Refs],Address,Dest) ->
  
%  references:ref_map(Refs,fun (Ref) -> 
  Block =  ref_adt:block(Ref),
  Offset =  ref_adt:offset(Ref),
  Place = Block + Offset*4,
  Type = ref_adt:type(Ref),
  NewDest = case Type of
	      rel -> Dest - Place;
	      abs -> Dest;
	      _ -> exit({bad_ref_type,Type})
	    end,
  Bits = ref_adt:bitlist(Ref),
  Instr = hipe_bifs:read_u32(Place),

  NewInst = patch_word(Bits,Instr,NewDest),

%  io:format("Patching ~w (~w[~w])\n from ~w to ~w by " ++
%	    "changing bits ~w\n",
%	    [Place,Block,Offset,Address,Dest,Bits]),

%  io:format("Patching ~w \n from    ",
%	    [Place]),
%  print_bits(Address bsr 16),
%  print_bits(Address band 16#ffff),  
%  {IHi,ILo} = Instr,
%  io:format("\nPatching ",[]),
%  print_bits(IHi),
%  print_bits(ILo),
%  io:format("\nto       ",[]),
%  print_bits(Hi),
%  print_bits(Lo),
%  io:format("\nto get   ",[]),
%  print_bits(NewDest bsr 16),
%  print_bits(NewDest band 16#ffff),
%  io:format(".\n",[]),

  hipe_bifs:write_u32(Place, NewInst),
  patch_refs(Refs,Address,Dest);
patch_refs([],_,_) -> true.

  
patch_word([],Instr,_) -> Instr;
patch_word([{From,To}|Rest],Instr,Dest) ->

  I = patch_word(From,To,Instr,Dest),
  patch_word(Rest,I,Dest).

patch_word(From,To,Instr,Dest) ->
  mask(To, Instr, get_bit(From,Dest)).

  
get_bit(From,Dest) -> (Dest bsr From) band 1.

mask(Pos,Word,Bit) -> (Word band (16#ffffffff - (1 bsl Pos))) bor (Bit bsl Pos).

print_bits(Word16) ->
  print_bits(15,Word16).

print_bits(0,Word) -> io:format("~w",[Word band 1]);
print_bits(N,Word) -> io:format("~w",[(Word band (1 bsl N)) bsr N]),
		      print_bits(N-1,Word).


