%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
%%----------------------------------------------------------------------
%% Purpose:
%%----------------------------------------------------------------------

-module(megaco_binary_term_id_test).

%%----------------------------------------------------------------------
%% Include files
%%----------------------------------------------------------------------
-include_lib("megaco/include/megaco.hrl"). 
-include_lib("megaco/src/engine/megaco_message_internal.hrl").


%%----------------------------------------------------------------------
%% External exports
%%----------------------------------------------------------------------
-export([t/0]).


%%----------------------------------------------------------------------
%% Internal exports
%%----------------------------------------------------------------------
-export([te01/1,te02/1,te03/1,te04/1,te05/1,
	 te06/1,te07/1,te08/1,te09/1,te10/1,
	 te11/1,te12/1,te13/1,te14/1,te15/1,
	 te16/1,te17/1,te18/1,te19/1]).
-export([td1/1,td2/1,td3/1,td4/1,td5/1,td6/1]).


%% ---------------------------------------------------------------------


%%======================================================================
%% External functions
%%======================================================================

t() ->
    display([do(Case) || Case <- cases()]),
    ok.


cases() -> encode_first_cases() ++ decode_first_cases().

encode_first_cases() -> [te01,te02,te03,te04,te05,
			 te06,te07,te08,te09,te10,
			 te11,te12,te13,te14,te15,
			 te16,te17,te18,te19].
decode_first_cases() -> [td1,td2,td3,td4,td5,td6].

do(Case) ->
    case doc(Case) of
	{'EXIT',_} ->
	    {Case,error};
	Description ->
	    io:format("test case ~p~n",[Case]),
	    case suite(Case) of
		{'EXIT',Reason} ->
		    Res = check_result(Case,Description,ok,{error,Reason}),
		    {Case,Res};
		{Expected,Result} ->
		    Res = check_result(Case,Description,Expected,Result),
		    {Case,Res}
	    end
    end.

doc(Case)   -> (catch apply(?MODULE,Case,[doc])).
suite(Case) -> (catch apply(?MODULE,Case,[suite])).

display(R) -> 
    [display(Case,Result) || {Case,Result} <- R].

display(C,error) ->
    io:format("Test case ~p failed~n",[C]);
display(C,warning) ->
    io:format("Test case ~p conspicuous~n",[C]);
display(C,ok) ->
    io:format("Test case ~p succeeded~n",[C]).

check_result(C,D,ok,{ok,T1,T2,T3}) ->
    Result = case check_ok_result(T1,T3) of
		 ok -> 
		     io:format("  => succeeded"
			       "~n  ~s"
			       "~n  ~p"
			       "~n  ~w"
			       "~n  ~p",
			       [D,T1,T2,T3]),
		     ok;
		 {error,Reason} ->
		     io:format("  => inconsistent result"
			       "~n  Start and end record differ"
			       "~n  ~s"
			       "~n  ~s"
			       "~n  ~w",
			       [D,Reason,T2]),
		     warning
	     end,
    io:format("~n~n--------------------~n",[]),
    Result;
check_result(C,D,error,{ok,T1,T2,T3}) ->
    io:format("  => failed"
	      "~n  ~s"
	      "~n  ~p"
	      "~n  ~p"
	      "~n  ~p"
	      "~n~n--------------------~n",
	      [D,T1,T2,T3]),
    error;
check_result(C,D,error,{error,Reason}) ->
    io:format("  => succeeded"
	      "~n  ~s"
	      "~n  Operation failed (expectedly) for reason"
	      "~n  ~p"
	      "~n~n--------------------~n",
	      [D,Reason]),
    ok;
check_result(C,D,ok,{error,Reason}) ->
    io:format("  => failed"
	      "~n  ~s"
	      "~n  Failed for reason"
	      "~n  ~p"
	      "~n~n--------------------~n",
	      [D,Reason]),
    error.

check_ok_result(R,R) when record(R,megaco_term_id) ->
    ok; % Same record type and same record content
check_ok_result(S,E) when record(S,megaco_term_id),
                          record(E,megaco_term_id) ->
    Reason = check_megaco_term_id_record(S,E),
    {error,Reason}; % Same record type but different record content
check_ok_result(R,R) when record(R,'TerminationID') ->
    ok;
check_ok_result(S,E) when record(S,'TerminationID'),
                          record(E,'TerminationID') ->
    Reason = check_TerminationID_record(S,E),
    {error,Reason}; % Same record type but different record content
check_ok_result(S,E) ->
    {error,"NOT THE SAME RECORD TYPES"}. % OOPS, Not even the same record type
    
check_megaco_term_id_record(#megaco_term_id{contains_wildcards = Cw1, 
					    id = Id1},
			    #megaco_term_id{contains_wildcards = Cw2, 
					    id = Id2}) ->
    Result = case check_megaco_term_id_cw(Cw1,Cw2) of
		 ok ->
		     check_megaco_term_id_id(Id1,Id2);
		 {error,R1} ->
		     R2 = check_megaco_term_id_id(Id1,Id2),
		     io_lib:format("~s~s",[R1,R2])
    end,
    lists:flatten(Result).
    
check_megaco_term_id_cw(Cw,Cw) ->
    ok;
check_megaco_term_id_cw(Cw1,Cw2) ->
    R = io_lib:format("~n   The 'contains_wildcard' property of the start"
		      "~n   megaco_term_id record "
		      "~n   has value ~w "
		      "~n   but the end record "
		      "~n   has value ~w",
		      [Cw1,Cw2]),
    {error,R}.

check_megaco_term_id_id(Id,Id) ->
    ok;
check_megaco_term_id_id(Id1,Id2) ->
    R = io_lib:format("~n   The 'id' property of the start"
		      "~n   megaco_term_id record "
		      "~n   has value ~w "
		      "~n   but the end record "
		      "~n   has value ~w",
		      [Id1,Id2]),
    {error,R}.


check_TerminationID_record(#'TerminationID'{wildcard = W1, id = Id1},
			   #'TerminationID'{wildcard = W2, id = Id2}) ->
    Result = case check_TerminationID_w(W1,W2) of
		 ok ->
		     check_TerminationID_id(Id1,Id2);
		 {error,R1} ->
		     R2 = check_TerminationID_id(Id1,Id2),
		     io_lib:format("~s~s",[R1,R2])
    end,
    lists:flatten(Result).
    
check_TerminationID_w(W,W) ->
    ok;
check_TerminationID_w(W1,W2) ->
    R = io_lib:format("~n   The 'wildcard' property of the start"
		      "~n   'TerminationID' record "
		      "~n   has value ~w "
		      "~n   but the end record "
		      "~n   has value ~w",
		      [W1,W2]),
    {error,R}.

check_TerminationID_id(Id,Id) ->
    ok;
check_TerminationID_id(Id1,Id2) ->
    R = io_lib:format("~n   The 'id' property of the start"
		      "~n   'TerminationID' record "
		      "~n   has value ~w "
		      "~n   but the end record "
		      "~n   has value ~w",
		      [Id1,Id2]),
    {error,R}.


%% --------------------------------------------------------
%% Start test cases
%% --------------------------------------------------------

te01(doc) ->
    "Basic encoding & then decoding test [1,1]\n  (asn -> binary -> asn)";

te01(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid01()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te02(doc) ->
    "Basic encoding & then decoding test [1,2]";

te02(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid02()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te03(doc) ->
    "Basic encoding & then decoding test [1,3]";

te03(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid03()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te04(doc) ->
    "Basic encoding & then decoding test [1,4]\n  (asn -> binary -> asn)";

te04(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid04()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te05(doc) ->
    "Basic encoding & then decoding test [1,5]\n  (asn -> binary -> asn)";

te05(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid05()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te06(doc) ->
    "Specified NO wildcards, but found some just the same [1,6]\n  (asn -> binary -> asn)";

te06(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid06()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te07(doc) ->
    "Invalid character found (2) [1,7]\n  (asn -> binary -> asn)";

te07(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid07()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te08(doc) ->
    "Erroneous length of first level (a character after wildcard) [1,8]\n  (asn -> binary -> asn)";

te08(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid08()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te09(doc) ->
    "Erroneous length of first level (a character after last valid) [1,9]\n  (asn -> binary -> asn)";

te09(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid09()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te10(doc) ->
    "Basic encode & decode with one recursive wildcard [1,10]\n  (asn -> binary -> asn)\n   (NOTE THAT THIS SHOULD LATER BE A LEVEL WILDCARD)";

te10(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid10()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te11(doc) ->
    "Basic encode & decode with one recursive wildcard [1,11]\n  (asn -> binary -> asn)";

te11(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid11()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te12(doc) ->
    "Basic encode & decode with one recursive wildcard [1,12]\n  (asn -> binary -> asn)";

te12(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid12()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te13(doc) ->
    "Basic encode & decode with one recursive wildcard [1,13]\n  (asn -> binary -> asn)";

te13(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid13()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te14(doc) ->
    "Basic encode & decode with one recursive wildcard [2,13]\n  (asn -> binary -> asn)";

te14(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config2(),mtid13()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te15(doc) ->
    "Basic encode & decode with one wildcard in the first level\nand then one empty level [1,15]\n  (asn -> binary -> asn)";

te15(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config1(),mtid15()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te16(doc) ->
    "Basic encode & decode with one recursive wildcard [2,15]\n  (asn -> binary -> asn)";

te16(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config2(),mtid15()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

te17(doc) ->
    "Basic encode & decode with one recursive wildcard [2,11]\n  (asn -> binary -> asn)";

te17(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config2(),mtid11()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te18(doc) ->
    "Basic encode & decode with one recursive wildcard [4,16]\n  (asn -> binary -> asn)";

te18(suite) ->
    put(encode_debug,dbg),
    put(decode_debug,dbg),
    Res = encode_decode(config4(),mtid16()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

te19(doc) ->
    "Basic encode & decode with one recursive wildcard [4,17]\n  (asn -> binary -> asn)";

te19(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = encode_decode(config4(),mtid17()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

td1(doc) ->
    "Root termination decoding & then encoding test [1,1]\n  (binary -> asn -> binary)";

td1(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config1(),atid1()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

%% One byte to much (should be two bytes but is three)
td2(doc) ->
    "Basic decoding & then encoding test [1,2]\n  (binary -> asn -> binary)";

td2(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config1(),atid2()),
    erase(encode_debug),
    erase(decode_debug),
    {error,Res}.


%% --------------------------------------------------------

td3(doc) ->
    "Basic decoding & then encoding test [2,2]\n  (binary -> asn -> binary)";

td3(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config2(),atid2()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

td4(doc) ->
    "Basic decoding & then encoding test [2,3]\n  (binary -> asn -> binary)";

td4(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config2(),atid3()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

td5(doc) ->
    "Basic decoding & then encoding test [2,4]\n  (binary -> asn -> binary)";

td5(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config2(),atid4()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------

td6(doc) ->
    "Basic decoding & then encoding test [3,5]\n  (binary -> asn -> binary)";

td6(suite) ->
    %% put(encode_debug,dbg),
    %% put(decode_debug,dbg),
    Res = decode_encode(config3(),atid5()),
    erase(encode_debug),
    erase(decode_debug),
    {ok,Res}.


%% --------------------------------------------------------
%% --------------------------------------------------------

encode_decode(C,T) ->
    case encode(C,T) of
	{ok,T1} ->
	    case decode(C,T1) of
		{ok,T2} ->
		    {ok,T,T1,T2};
		{error,R2} ->
		    {error,{decode_error,T,T1,R2}};
		{exit,E2} ->
		    {error,{decode_exit,T,T1,E2}}
	    end;
	{error,R1} ->
	    {error,{encode_error,T,R1}};
	{exit,E1} ->
	    {error,{encode_exit,T,E1}}
    end.

decode_encode(C,T) ->
    case decode(C,T) of
	{ok,T1} ->
	    case encode(C,T1) of
		{ok,T2} ->
		    {ok,T,T1,T2};
		{error,R2} ->
		    {error,{encode_error,T,T1,R2}};
		{exit,E2} ->
		    {error,{encode_exit,T,T1,E2}}
	    end;
	{error,R1} ->
	    {error,{decode_error,T,R1}};
	{exit,E1} ->
	    {error,{decode_exit,T,E1}}
    end.

%% ------------------

config1() -> [3,8,5].
config2() -> [3,5,4,8,4].
config3() -> [3,5,4,16,4].
config4() -> [8,8,8].  % Default config

mtid01() -> 
    #megaco_term_id{contains_wildcards = false,
		    id = [[$1,$0,$1],
			  [$1,$1,$0,$1,$0,$0,$0,$1],
			  [$1,$1,$0,$1,$1]]}.
mtid02() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$$],[$1,$1,$0,$1,$$],[$1,$1,$*]]}.
mtid03() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$1],[$1,$0,$1,$$],[$1,$1,$*]]}.
mtid04() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$$],[$1,$1,$0,$1,$$],[$1,$1,$1,$0,$0]]}.
mtid05() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$1],[$1,$1,$0,$1,$$],[$1,$1,$1,$0,$0]]}.
mtid06() -> 
    #megaco_term_id{contains_wildcards = false,
		    id = [[$1,$0,$$],[$1,$1,$0,$1,$$],[$1,$1,$*]]}.
mtid07() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$*],[$1,2,$0,$1,$$],[$1,$1,$1,$0,$0]]}.
mtid08() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$*,$1],[$1,$1,$0,$1,$$],[$1,$1,$1,$0,$0]]}.
mtid09() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$0,$1],[$1,$1,$0,$1,$$],[$1,$1,$1,$0,$0]]}.
mtid10() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$1],[$1,$0,$1,$$],[$$]]}.
mtid11() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$1],[$1,$0,$1,$$]]}.
mtid12() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$1],[$1,$0,$1,$0,$1,$1,$$]]}.
mtid13() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$*]]}.
%% Empty last level
mtid14() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$*],[$1,$0,$0,$$],[]]}.
%% Empty second level and missing last level
mtid15() -> 
    #megaco_term_id{contains_wildcards = true,
		    id = [[$1,$0,$*],[]]}.
%% Megaco all wildcard termination id
mtid16() ->
	#megaco_term_id{contains_wildcards = true,
			id                 = [[?megaco_all]]}.
%% Megaco choose wildcard termination id
mtid17() ->
	#megaco_term_id{contains_wildcards = true,
			id                 = [[?megaco_choose]]}.

aroot() ->
    #'TerminationID'{wildcard = [],
		     id       = [16#FF, 16#FF, 16#FF, 16#FF,
				 16#FF, 16#FF, 16#FF, 16#FF]}.
atid1() -> aroot().
atid2() ->
    #'TerminationID'{wildcard = [],
		     id       = [2#00000001, 02#00011110, 2#0]}.
atid3() ->
    #'TerminationID'{wildcard = [[2#00010111], [2#00000111]],
		     id       = [2#0, 2#00011110, 2#0]}.
atid4() -> 
    #'TerminationID'{wildcard = [[2#01001111]],
		     id       = [2#0000001, 2#0, 2#0]}.
%% 
atid5() -> 
    #'TerminationID'{wildcard = [[2#01001111]],
		     id       = [2#00000001, 2#00110011, 
				 2#00000000, 2#00000000]}.

%%--

    
%% ------------------

encode(C,T) ->
    encode(get(encode_debug),C,T).

encode(L,C,T) -> 
    put(dsev,L),
    Res = encode1(C,T),
    erase(dsev),
    Res;
encode(_,C,T) ->
    encode1(C,T).

encode1(C,T) ->
    case (catch megaco_binary_term_id:encode(C,T)) of
	{'EXIT',Reason} ->
	    {exit,Reason};
	Else ->
	    Else
    end.

decode(C,T) ->
    decode(get(decode_debug),C,T).

decode(L,C,T) ->
    put(dsev,L),
    Res = decode1(C,T),
    erase(dsev),
    Res;
decode(_,C,T) ->
    decode1(C,T).

decode1(C,T) -> 
    case (catch megaco_binary_term_id:decode(C,T)) of
	{'EXIT',Reason} ->
	    {exit,Reason};
	Else ->
	    Else
    end.




